//-----------------------------------------------------------------------------
// File: Meshes.cpp
//
// Desc: For advanced geometry, most apps will prefer to load pre-authored
//       Meshes from a file. Fortunately, when using Meshes, D3DX does most of
//       the work for this, parsing a geometry file and creating vertx buffers
//       (and index buffers) for us. This tutorial shows how to use a D3DXMESH
//       object, including loading it from a file and rendering it. One thing
//       D3DX does not handle for us is the materials and textures for a mesh,
//       so note that we have to handle those manually.
//
//       Note: one advanced (but nice) feature that we don't show here is that
//       when cloning a mesh we can specify the FVF. So, regardless of how the
//       mesh was authored, we can add/remove normals, add more texture
//       coordinate sets (for multi-texturing), etc.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------
#include <Windows.h>
#include <mmsystem.h>
#include <d3dx9.h>
#pragma warning( disable : 4996 ) // disable deprecated warning 
#include <strsafe.h>
#pragma warning( default : 4996 ) 

#include <d3dx9mesh.h>

// PSgȂꍇ0A
// gꍇ1ݒ肵āAhlsl.fxtechnique ShadeShader̃RgAEgO
#define ENABLE_PS 0

//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------
LPDIRECT3D9             g_pD3D           = NULL; // Used to create the D3DDevice
LPDIRECT3DDEVICE9       g_pd3dDevice     = NULL; // Our rendering device

LPD3DXMESH              g_pMesh          = NULL; // Our mesh object in sysmem
D3DMATERIAL9*           g_pMeshMaterials = NULL; // Materials for our mesh
LPDIRECT3DTEXTURE9*     g_pMeshTextures  = NULL; // Textures for our mesh
DWORD                   g_dwNumMaterials = 0L;   // Number of mesh materials

D3DXMATRIXA16 g_MatProj;
D3DXMATRIXA16 g_MatView;
D3DXMATRIXA16 g_MatWorld;

LPDIRECT3DTEXTURE9 g_pShadeTexture  = NULL;
LPDIRECT3DTEXTURE9 g_pWhiteTexture  = NULL;
LPD3DXEFFECT g_pEffect = NULL;


//-----------------------------------------------------------------------------
// Name: InitD3D()
// Desc: Initializes Direct3D
//-----------------------------------------------------------------------------
HRESULT InitD3D( HWND hWnd )
{
    // Create the D3D object.
    if( NULL == ( g_pD3D = Direct3DCreate9( D3D_SDK_VERSION ) ) )
        return E_FAIL;

    // Set up the structure used to create the D3DDevice. Since we are now
    // using more complex geometry, we will create a device with a zbuffer.
    D3DPRESENT_PARAMETERS d3dpp; 
    ZeroMemory( &d3dpp, sizeof(d3dpp) );
    d3dpp.Windowed = TRUE;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
    d3dpp.EnableAutoDepthStencil = TRUE;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D16;

	// o΃A`GCAX
	if( MessageBox( NULL, "A`GCAXONɂ܂H", "ݒ", MB_YESNO ) == IDYES ){
		d3dpp.MultiSampleType = D3DMULTISAMPLE_4_SAMPLES;
	}

    // Create the D3DDevice
    if( FAILED( g_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
                                      D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                                      &d3dpp, &g_pd3dDevice ) ) )
    {
        return E_FAIL;
    }

    g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE );
    g_pd3dDevice->SetRenderState( D3DRS_LIGHTING, FALSE );

	LPD3DXBUFFER pCompilationErrors;
    if( FAILED ( 
		D3DXCreateEffectFromFile( g_pd3dDevice, "hlsl.fx", NULL, NULL, D3DXSHADER_DEBUG | D3DXSHADER_SKIPOPTIMIZATION, NULL, &g_pEffect, &pCompilationErrors ) )
	){
		MessageBox( NULL, (LPCSTR)pCompilationErrors->GetBufferPointer(), "Error", MB_OK );
        return E_FAIL;
	}

	g_pEffect->OnResetDevice();

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: InitGeometry()
// Desc: Load the mesh and build the material and texture arrays
//-----------------------------------------------------------------------------
HRESULT InitGeometry()
{
    LPD3DXBUFFER pD3DXMtrlBuffer;

    // Load the mesh from the specified file
    if( FAILED( D3DXLoadMeshFromX( ".\\data\\xfile\\test.x", D3DXMESH_SYSTEMMEM, 
                                   g_pd3dDevice, NULL, 
                                   &pD3DXMtrlBuffer, NULL, &g_dwNumMaterials, 
                                   &g_pMesh ) ) )
    {
        MessageBox(NULL, "Could not find test.x", "Meshes.exe", MB_OK);
        return E_FAIL;
    }

    // We need to extract the material properties and texture names from the 
    // pD3DXMtrlBuffer
    D3DXMATERIAL* d3dxMaterials = (D3DXMATERIAL*)pD3DXMtrlBuffer->GetBufferPointer();
    g_pMeshMaterials = new D3DMATERIAL9[g_dwNumMaterials];
    if( g_pMeshMaterials == NULL )
        return E_OUTOFMEMORY;
    g_pMeshTextures  = new LPDIRECT3DTEXTURE9[g_dwNumMaterials];
    if( g_pMeshTextures == NULL )
        return E_OUTOFMEMORY;

    for( DWORD i=0; i<g_dwNumMaterials; i++ )
    {
        // Copy the material
        g_pMeshMaterials[i] = d3dxMaterials[i].MatD3D;

        // Set the ambient color for the material (D3DX does not do this)
        g_pMeshMaterials[i].Ambient = g_pMeshMaterials[i].Diffuse;

        g_pMeshTextures[i] = NULL;
        if( d3dxMaterials[i].pTextureFilename != NULL && 
            lstrlen(d3dxMaterials[i].pTextureFilename) > 0 )
        {
            // Create the texture
            if( FAILED( D3DXCreateTextureFromFile( g_pd3dDevice, 
                                                d3dxMaterials[i].pTextureFilename, 
                                                &g_pMeshTextures[i] ) ) )
            {
                MessageBox(NULL, "Could not find texture map", "Meshes.exe", MB_OK);
            }
        }
    }

    // Done with the material buffer
    pD3DXMtrlBuffer->Release();

	// eeNX`ǂݍ
	D3DXCreateTextureFromFile( g_pd3dDevice, "shade.bmp", &g_pShadeTexture );

	// eNX`(PSp)
	// fɃeNX`ĂȂꍇłAPSŋIɃeNX`FgĂ̂ŁA
	// _~[̃eNX`pӂĂB
	// nVIDIAHAL̏ꍇSetTextureNULLw肵Ă΁A
	// eNX`gȂĂƓĂ邪A
	// REFƃeNX`FɂȂĂ܂̂ŔÔ߁B
	// Ả@CB
#if ENABLE_PS == 1
	D3DXCreateTexture( g_pd3dDevice, 1, 1, 0, 0, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &g_pWhiteTexture );
	D3DLOCKED_RECT Rect;
	g_pWhiteTexture->LockRect( 0, &Rect, NULL, 0 );
	*(DWORD *)Rect.pBits = 0xFFFFFFFF;
	g_pWhiteTexture->UnlockRect( 0 );
#endif

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: Cleanup()
// Desc: Releases all previously initialized objects
//-----------------------------------------------------------------------------
VOID Cleanup()
{
    if( g_pMeshMaterials != NULL ) 
        delete[] g_pMeshMaterials;

    if( g_pMeshTextures )
    {
        for( DWORD i = 0; i < g_dwNumMaterials; i++ )
        {
            if( g_pMeshTextures[i] )
                g_pMeshTextures[i]->Release();
        }
        delete[] g_pMeshTextures;
    }
    if( g_pMesh != NULL )
        g_pMesh->Release();
    
    if( g_pd3dDevice != NULL )
        g_pd3dDevice->Release();

    if( g_pD3D != NULL )
        g_pD3D->Release();
}



//-----------------------------------------------------------------------------
// Name: SetupMatrices()
// Desc: Sets up the world, view, and projection transform matrices.
//-----------------------------------------------------------------------------
VOID SetupMatrices()
{
    // Set up world matrix
    D3DXMatrixRotationY( &g_MatWorld, timeGetTime() / 1000.0f );

    // Set up our view matrix. A view matrix can be defined given an eye point,
    // a point to lookat, and a direction for which way is up. Here, we set the
    // eye five units back along the z-axis and up three units, look at the 
    // origin, and define "up" to be in the y-direction.
    D3DXVECTOR3 vEyePt( 0.0f, 0.5f, -1.5f );
    D3DXVECTOR3 vLookatPt( 0.0f, 0.0f, 0.0f );
    D3DXVECTOR3 vUpVec( 0.0f, 1.0f, 0.0f );
    D3DXMatrixLookAtLH( &g_MatView, &vEyePt, &vLookatPt, &vUpVec );

    // For the projection matrix, we set up a perspective transform (which
    // transforms geometry from 3D view space to 2D viewport space, with
    // a perspective divide making objects smaller in the distance). To build
    // a perpsective transform, we need the field of view (1/4 pi is common),
    // the aspect ratio, and the near and far clipping planes (which define at
    // what distances geometry should be no longer be rendered).
    D3DXMatrixPerspectiveFovLH( &g_MatProj, D3DX_PI/4, 1.0f, 1.0f, 100.0f );
}




//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Draws the scene
//-----------------------------------------------------------------------------
VOID Render()
{
    // Clear the backbuffer and the zbuffer
    g_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, 
                         D3DCOLOR_XRGB(0,0,255), 1.0f, 0 );
    
    // Begin the scene
    if( SUCCEEDED( g_pd3dDevice->BeginScene() ) )
    {
        // Setup the world, view, and projection matrices
        SetupMatrices();

        UINT nPasses;
        UINT iPass;

		// gD[VF[_
		g_pEffect->SetTechnique( g_pEffect->GetTechniqueByName( "ShadeShader" ) );
		
        g_pEffect->SetMatrix( "MatrixW", &g_MatWorld );

		D3DXMATRIX m = g_MatWorld * g_MatView * g_MatProj;
        g_pEffect->SetMatrix( "MatrixWVP", &m );

		g_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );

#if ENABLE_PS == 0
	    // Œ蒸_gꍇ
		// eNX`Xe[W0 ݒ
		g_pd3dDevice->SetSamplerState( 0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );
		g_pd3dDevice->SetSamplerState( 0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );
		g_pd3dDevice->SetSamplerState( 0, D3DSAMP_MINFILTER, D3DSAMP_MIPFILTER );

		g_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_DIFFUSE );
		g_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_TEXTURE );
		g_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_MODULATE );

		// eNX`Xe[W1 ݒ
		g_pd3dDevice->SetSamplerState( 1, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );
		g_pd3dDevice->SetSamplerState( 1, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );
		g_pd3dDevice->SetSamplerState( 1, D3DSAMP_MINFILTER, D3DSAMP_MIPFILTER );

		g_pd3dDevice->SetTextureStageState( 1, D3DTSS_COLORARG1, D3DTA_CURRENT );
		g_pd3dDevice->SetTextureStageState( 1, D3DTSS_COLORARG2, D3DTA_TEXTURE );
		g_pd3dDevice->SetTextureStageState( 1, D3DTSS_COLOROP, D3DTOP_MODULATE );
#endif
		g_pd3dDevice->SetSamplerState( 1, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP );
		g_pd3dDevice->SetSamplerState( 1, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP );

		g_pEffect->Begin( &nPasses, 0 );

		// Cg̐ݒ
		D3DCOLORVALUE LightDiffuseColor	= { 1.0f, 1.0f, 1.0f, 1.0f };
		D3DCOLORVALUE LightAmbientColor	= { 0.0f, 0.0f, 0.0f, 1.0f };
		g_pEffect->SetValue( "LightDiffuseColor", &LightDiffuseColor, sizeof( D3DCOLORVALUE ) );
		g_pEffect->SetValue( "LightAmbientColor", &LightAmbientColor, sizeof( D3DCOLORVALUE ) );

		D3DXVECTOR3 LightDirection = D3DXVECTOR3( 1.0f, -0.5f, 1.0f ); // [hW
		D3DXVec3Normalize ( &LightDirection, &LightDirection );

		g_pEffect->SetValue( "LightDirection", &LightDirection, sizeof( D3DXVECTOR3 ) );

		for( iPass = 0; iPass < nPasses; iPass ++ ){
			g_pEffect->BeginPass( iPass );

			// Meshes are divided into subsets, one for each material. Render them in
			// a loop
			for( DWORD i=0; i<g_dwNumMaterials; i++ )
			{
				// Set the material and texture for this subset

#if ENABLE_PS == 0
				 // Œ蒸_gꍇ
				g_pd3dDevice->SetTexture( 0, g_pMeshTextures[i] );
				g_pd3dDevice->SetTexture( 1, g_pShadeTexture );

				if( g_pMeshTextures[i] ) g_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_MODULATE );
				else g_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_SELECTARG1 );
#else
				// PSgꍇ
				g_pEffect->SetTexture( "ShadeTexture", g_pShadeTexture );
				if( g_pMeshTextures[i] ) g_pEffect->SetTexture( "MeshTexture", g_pMeshTextures[i] );
				else g_pEffect->SetTexture( "MeshTexture", g_pWhiteTexture );
#endif

				g_pEffect->SetValue( "MaterialAmbientColor", &g_pMeshMaterials[i].Ambient, sizeof( D3DCOLORVALUE ) );
				g_pEffect->SetValue( "MaterialDiffuseColor", &g_pMeshMaterials[i].Diffuse, sizeof( D3DCOLORVALUE ) );

				g_pEffect->CommitChanges();

				// Draw the mesh subset
				g_pMesh->DrawSubset( i );
			}

			g_pEffect->EndPass();
		}
		g_pEffect->End();

		// ֊sVF[_
		g_pEffect->SetTechnique( g_pEffect->GetTechniqueByName( "OutlineShader" ) );

		g_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_CW );
		g_pd3dDevice->SetTexture( 0, NULL );
		g_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1 );
		g_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TFACTOR );
		g_pd3dDevice->SetRenderState( D3DRS_TEXTUREFACTOR, 0xFF400000 ); // ֊s̐F

		g_pEffect->Begin( &nPasses, 0 );
		for( iPass = 0; iPass < nPasses; iPass ++ ){
			g_pEffect->BeginPass( iPass );

			// Meshes are divided into subsets, one for each material. Render them in
			// a loop
			for( DWORD i=0; i<g_dwNumMaterials; i++ )
			{
				// Draw the mesh subset
				g_pMesh->DrawSubset( i );
			}

			g_pEffect->EndPass();
		}
		g_pEffect->End();


        // End the scene
        g_pd3dDevice->EndScene();
    }

    // Present the backbuffer contents to the display
    g_pd3dDevice->Present( NULL, NULL, NULL, NULL );
}




//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: The window's message handler
//-----------------------------------------------------------------------------
LRESULT WINAPI MsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
    switch( msg )
    {
        case WM_DESTROY:
            Cleanup();
            PostQuitMessage( 0 );
            return 0;
    }

    return DefWindowProc( hWnd, msg, wParam, lParam );
}




//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: The application's entry point
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
    // Register the window class
    WNDCLASSEX wc = { sizeof(WNDCLASSEX), CS_CLASSDC, MsgProc, 0L, 0L, 
                      GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
                      "D3D Tutorial", NULL };
    RegisterClassEx( &wc );

    // Create the application's window
    HWND hWnd = CreateWindow( "D3D Tutorial", "D3D Tutorial 06: Meshes", 
                              WS_OVERLAPPEDWINDOW, 100, 100, 512, 512,
                              NULL, NULL, wc.hInstance, NULL );

    // Initialize Direct3D
    if( SUCCEEDED( InitD3D( hWnd ) ) )
    { 
        // Create the scene geometry
        if( SUCCEEDED( InitGeometry() ) )
        {
            // Show the window
            ShowWindow( hWnd, SW_SHOWDEFAULT );
            UpdateWindow( hWnd );

            // Enter the message loop
            MSG msg; 
            ZeroMemory( &msg, sizeof(msg) );
            while( msg.message!=WM_QUIT )
            {
                if( PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE ) )
                {
                    TranslateMessage( &msg );
                    DispatchMessage( &msg );
                }
                else
                    Render();
            }
        }
    }

    UnregisterClass( "D3D Tutorial", wc.hInstance );
    return 0;
}



